import 'dart:convert';

import 'package:ink_runtime/ink.dart';
import 'package:test/test.dart';

import '../../base_ink_tests.dart';
import '../inkfiles/compiled/inkjs/tests.ink.json.dart' as $tests;

class InkjsChoicesSpec extends BaseInkTests {
  InkjsChoicesSpec(Future<Story> Function(String) load,
      {required TestFunction test, required GroupFunction group})
      : super(
          spec: 'Choices',
          createStory: load,
          groupFunction: group,
          testFunction: test,
        );

  late Story story;

  @override
  void tests() {
    setUp(() async {
      story = await createStory(jsonEncode($tests.json));
    });

    it('should offer a single choice', () {
      story.ChoosePathString('choices.basic_choice');

      story.Continue();
      expect(story.currentChoices.length, 1);
      expect(story.canContinue, false);
    });

    it('should offer multiple choices', () {
      story.ChoosePathString('choices.multiple_choices');

      story.Continue();
      expect(story.currentChoices.length, 3);
      expect(story.canContinue, false);
    });

    it('should select a choice', () {
      story.ChoosePathString('choices.multiple_choices');

      story.Continue();
      story.ChooseChoiceIndex(0);
      expect(story.Continue(), 'choice 1\n');
      expect(story.canContinue, false);
    });

    it('should throw when selecting an invalid choice', () {
      story.ChoosePathString('choices.multiple_choices');

      story.Continue();
      expect(() => story.ChooseChoiceIndex(10), throwsA(anything));
    });

    it('should suppress parts of choice text', () {
      story.ChoosePathString('choices.choice_text');

      story.Continue();
      expect(story.currentChoices.length, 1);
      expect(story.canContinue, false);

      expect(story.currentChoices[0].text, 'always choice only');
      story.ChooseChoiceIndex(0);
      expect(story.canContinue, true);
      expect(story.Continue(), 'always output only\n');
      expect(story.canContinue, false);
    });

    it('should suppress choices after they have been selected', () {
      story.ChoosePathString('choices.suppression');
      expect(story.canContinue, true);
      story.Continue();

      expect(story.currentChoices.length, 2);
      expect(story.currentChoices[0].text, 'choice 1');
      expect(story.currentChoices[1].text, 'choice 2');

      story.ChooseChoiceIndex(1);
      expect(story.Continue(), 'choice 2\n');
      expect(story.canContinue, false);

      story.ChoosePathString('choices.suppression');
      expect(story.canContinue, true);
      story.Continue();

      expect(story.currentChoices.length, 1);
      expect(story.currentChoices[0].text, 'choice 1');

      story.ChooseChoiceIndex(0);
      expect(story.Continue(), 'choice 1\n');
      expect(story.canContinue, false);

      story.ChoosePathString('choices.suppression');
      expect(story.canContinue, true);
      // TODO test for exception
    });

    it('should select the fallback choice', () {
      story.ChoosePathString('choices.fallback');
      expect(story.canContinue, true);
      story.Continue();

      expect(story.currentChoices.length, 1);
      expect(story.currentChoices[0].text, 'choice 1');
      story.ChooseChoiceIndex(0);
      story.Continue();

      story.ChoosePathString('choices.fallback');
      expect(story.canContinue, true);
      story.Continue();

      expect(story.currentChoices.length, 0);
      expect(story.canContinue, false);
    });

    it('should keep a sticky choice', () {
      story.ChoosePathString('choices.sticky');
      expect(story.canContinue, true);
      story.Continue();

      expect(story.currentChoices.length, 2);
      expect(story.currentChoices[0].text, 'disapears');
      expect(story.currentChoices[1].text, 'stays');

      story.ChooseChoiceIndex(0);
      story.Continue();

      for (var i = 0; i < 3; ++i) {
        story.ChoosePathString('choices.sticky');
        expect(story.canContinue, true);
        story.Continue();

        expect(story.currentChoices.length, 1);
        expect(story.currentChoices[0].text, 'stays');
        story.ChooseChoiceIndex(0);
        expect(story.Continue(), 'stays\n');
      }
    });

    it('should handle conditional choices', () {
      story.ChoosePathString('choices.conditional');
      expect(story.canContinue, true);
      story.Continue();

      expect(story.currentChoices.length, 3);
      expect(story.currentChoices[0].text, 'no condition');
      expect(story.currentChoices[1].text, 'available');
      expect(story.currentChoices[2].text, 'multi condition available');
    });
  }
}
